@@ -4,11 +4,12 @@
  *	This module implements cell oriented functions for table
  *	widgets.
  *
- * Copyright (c) 1998-1999 Jeffrey Hobbs
+ * Copyright (c) 1998-2000 Jeffrey Hobbs
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
+ * RCS: @(#) $Id: tkTableCell.c,v 1.17 2004/02/08 03:09:46 cerney Exp $
  */
  
 #include "tkVMacro.h"
@@ -39,7 +40,8 @@ int
 TableTrueCell(Table *tablePtr, int r, int c, int *row, int *col)
 {
     *row = r; *col = c;
-    /* We check spans before constraints, because we don't want to
+    /*
+     * We check spans before constraints, because we don't want to
      * constrain and then think we ended up in a span
      */
     if (tablePtr->spanAffTbl && !(tablePtr->flags & AVOID_SPANS)) {
@@ -50,8 +52,10 @@ TableTrueCell(Table *tablePtr, int r, int c, int *row, int *col)
 	entryPtr = Tcl_FindHashEntry(tablePtr->spanAffTbl, buf);
 	if ((entryPtr != NULL) &&
 		((char *)Tcl_GetHashValue(entryPtr) != NULL)) {
-	    /* This cell is covered by another spanning cell */
-	    /* We need to return the coords for that cell */
+	    /*
+	     * This cell is covered by another spanning cell.
+	     * We need to return the coords for that spanning cell.
+	     */
 	    TableParseArrayIndex(row, col, (char *)Tcl_GetHashValue(entryPtr));
 	    return 0;
 	}
@@ -94,11 +98,8 @@ TableCellCoords(Table *tablePtr, int row, int col,
      * Real coords required, always should be passed acceptable values,
      * but this is a possible seg fault otherwise
      */
-#define BOUNDS_CHECK 1
-#ifdef BOUNDS_CHECK
     CONSTRAIN(row, 0, tablePtr->rows-1);
     CONSTRAIN(col, 0, tablePtr->cols-1);
-#endif
     *w = tablePtr->colPixels[col];
     *h = tablePtr->rowPixels[row];
     /*
@@ -131,8 +132,11 @@ TableCellCoords(Table *tablePtr, int row, int col,
 	    cell = (char *) Tcl_GetHashValue(entryPtr);
 	    TableParseArrayIndex(&rs, &cs, cell);
 	    if (rs > 0) {
+		/*
+		 * Make sure we don't overflow our space
+		 */
 		if (row < tablePtr->titleRows) {
-		    rs = MIN(tablePtr->titleRows-1-row, rs);
+		    rs = MIN(tablePtr->titleRows-1, row+rs);
 		} else {
 		    rs = MIN(tablePtr->rows-1, row+rs);
 		}
@@ -142,8 +146,11 @@ TableCellCoords(Table *tablePtr, int row, int col,
 		/* currently negative spans are not supported */
 	    }
 	    if (cs > 0) {
+		/*
+		 * Make sure we don't overflow our space
+		 */
 		if (col < tablePtr->titleCols) {
-		    cs = MIN(tablePtr->titleCols-1-col, cs);
+		    cs = MIN(tablePtr->titleCols-1, col+cs);
 		} else {
 		    cs = MIN(tablePtr->cols-1, col+cs);
 		}
@@ -194,19 +201,25 @@ TableCellVCoords(Table *tablePtr, int row, int col,
 
     if (tablePtr->tkwin == NULL) return 0;
 
-    /* Necessary to use separate vars in case dummies are passed in */
+    /*
+     * Necessary to use separate vars in case dummies are passed in
+     */
     cellType = TableCellCoords(tablePtr, row, col, &x, &y, &w, &h);
     *rx = x; *ry = y; *rw = w; *rh = h;
     if (cellType == CELL_OK) {
 	if ((row < tablePtr->topRow && row >= tablePtr->titleRows) ||
 	    (col < tablePtr->leftCol && col >= tablePtr->titleCols)) {
-	    /* non-spanning cell hiding in "dead" space
-	     * between title areas and visible cells */
+	    /*
+	     * A non-spanning cell hiding in "dead" space
+	     * between title areas and visible cells
+	     */
 	    return 0;
 	}
     } else if (cellType == CELL_SPAN) {
-	/* we might need to treat full better is CELL_SPAN but primary
-	 * cell is visible */
+	/*
+	 * we might need to treat full better is CELL_SPAN but primary
+	 * cell is visible
+	 */
 	int topX = tablePtr->colStarts[tablePtr->titleCols]+hl;
 	int topY = tablePtr->rowStarts[tablePtr->titleRows]+hl;
 	if ((col < tablePtr->leftCol) && (col >= tablePtr->titleCols)) {
@@ -225,26 +238,35 @@ TableCellVCoords(Table *tablePtr, int row, int col,
 		y = topY;
 	    }
 	}
-	/* re-set these according to changed coords */
+	/*
+	 * re-set these according to changed coords
+	 */
 	*rx = x; *ry = y; *rw = w; *rh = h;
     } else {
-	/* If it is a hidden cell, then w,h is the row,col in user coords
-	 * of the cell that spans over this one */
+	/*
+	 * If it is a hidden cell, then w,h is the row,col in user coords
+	 * of the cell that spans over this one
+	 */
 	return 0;
     }
-    /* At this point, we know it is on the screen,
+    /*
+     * At this point, we know it is on the screen,
      * but not if we can see 100% of it (if we care)
      */
     if (full) {
 	w0 = w; h0 = h;
     } else {
-	/* if we don't care about seeing the whole thing, then
-	 * make sure we at least see a pixel worth */
+	/*
+	 * if we don't care about seeing the whole thing, then
+	 * make sure we at least see a pixel worth
+	 */
 	w0 = h0 = 1;
     }
-    /* Is the cell visible? */
-    if (x<hl || y<hl || (x+w0)>(Tk_Width(tablePtr->tkwin)-hl)
-	|| (y+h0)>(Tk_Height(tablePtr->tkwin)-hl)) {
+    /*
+     * Is the cell visible?
+     */
+    if ((x < hl) || (y < hl) || ((x+w0) > (Tk_Width(tablePtr->tkwin)-hl))
+	    || ((y+h0) > (Tk_Height(tablePtr->tkwin)-hl))) {
 	/* definitely off the screen */
 	return 0;
     } else {
@@ -335,18 +357,26 @@ TableWhatCell(register Table *tablePtr, int x, int y, int *row, int *col)
 int
 TableAtBorder(Table * tablePtr, int x, int y, int *row, int *col)
 {
-    int i, brow, bcol, borders = 2, bd = tablePtr->defaultTag.bd;
-    int dbd = 2*bd;
+    int i, brow, bcol, borders = 2, bd[6];
+
+    TableGetTagBorders(&(tablePtr->defaultTag),
+	    &bd[0], &bd[1], &bd[2], &bd[3]);
+    bd[4] = (bd[0] + bd[1])/2;
+    bd[5] = (bd[2] + bd[3])/2;
+
+    /*
+     * Constrain x && y appropriately, and adjust x if it is not in the
+     * column titles to change display coords into internal coords.
+     */
     x = MAX(0, x); y = MAX(0, y);
     x -= tablePtr->highlightWidth; y -= tablePtr->highlightWidth;
-    /* Adjust the x coord if not in the column titles to change display
-     * coords into internal coords */
     x += (x < tablePtr->colStarts[tablePtr->titleCols]) ? 0 :
 	tablePtr->colStarts[tablePtr->leftCol] -
 	tablePtr->colStarts[tablePtr->titleCols];
     x = MIN(x, tablePtr->maxWidth - 1);
-    for (i = 1; i <= tablePtr->cols && x+dbd >= tablePtr->colStarts[i]; i++);
-    if (x > tablePtr->colStarts[--i]+bd) {
+    for (i = 1; (i <= tablePtr->cols) &&
+	     (x + (bd[0] + bd[1])) >= tablePtr->colStarts[i]; i++);
+    if (x > tablePtr->colStarts[--i] + bd[4]) {
 	borders--;
 	*col = -1;
 	bcol = (i < tablePtr->leftCol && i >= tablePtr->titleCols) ?
@@ -359,8 +389,9 @@ TableAtBorder(Table * tablePtr, int x, int y, int *row, int *col)
 	tablePtr->rowStarts[tablePtr->topRow] -
 	tablePtr->rowStarts[tablePtr->titleRows];
     y = MIN(y, tablePtr->maxHeight - 1);
-    for (i = 1; i <= tablePtr->rows && y+dbd >= tablePtr->rowStarts[i]; i++);
-    if (y > tablePtr->rowStarts[--i]+bd) {
+    for (i = 1; i <= tablePtr->rows &&
+	     (y + (bd[2] + bd[3])) >= tablePtr->rowStarts[i]; i++);
+    if (y > tablePtr->rowStarts[--i]+bd[5]) {
 	borders--;
 	*row = -1;
 	brow = (i < tablePtr->topRow && i >= tablePtr->titleRows) ?
@@ -369,20 +400,33 @@ TableAtBorder(Table * tablePtr, int x, int y, int *row, int *col)
 	brow = *row = (i < tablePtr->topRow && i >= tablePtr->titleRows) ?
 	    tablePtr->titleRows-1 : i-1;
     }
-    /* Have to account for spanning cells */
+    /*
+     * We have to account for spanning cells, which may hide cells.
+     * In that case, we have to decrement our border count.
+     */
     if (tablePtr->spanAffTbl && !(tablePtr->flags & AVOID_SPANS) && borders) {
+	Tcl_HashEntry *entryPtr1, *entryPtr2 ;
 	char buf1[INDEX_BUFSIZE], buf2[INDEX_BUFSIZE];
+	char *val;
 
-	/* Have to account for "hidden" cells */
 	if (*row != -1) {
 	    TableMakeArrayIndex(brow+tablePtr->rowOffset,
 				bcol+tablePtr->colOffset+1, buf1);
 	    TableMakeArrayIndex(brow+tablePtr->rowOffset+1,
 				bcol+tablePtr->colOffset+1, buf2);
-	    if (Tcl_FindHashEntry(tablePtr->spanAffTbl, buf1) != NULL &&
-		Tcl_FindHashEntry(tablePtr->spanAffTbl, buf2) != NULL) {
-		borders--;
-		*row = -1;
+	    entryPtr1 = Tcl_FindHashEntry(tablePtr->spanAffTbl, buf1);
+	    entryPtr2 = Tcl_FindHashEntry(tablePtr->spanAffTbl, buf2);
+	    if (entryPtr1 != NULL && entryPtr2 != NULL) {
+		if ((val = (char *) Tcl_GetHashValue(entryPtr1)) != NULL) {
+		    strcpy(buf1, val);
+		}
+		if ((val = (char *) Tcl_GetHashValue(entryPtr2)) != NULL) {
+		    strcpy(buf2, val);
+		}
+		if (strcmp(buf1, buf2) == 0) {
+		    borders--;
+		    *row = -1;
+		}
 	    }
 	}
 	if (*col != -1) {
@@ -390,10 +434,19 @@ TableAtBorder(Table * tablePtr, int x, int y, int *row, int *col)
 				bcol+tablePtr->colOffset, buf1);
 	    TableMakeArrayIndex(brow+tablePtr->rowOffset+1,
 				bcol+tablePtr->colOffset+1, buf2);
-	    if (Tcl_FindHashEntry(tablePtr->spanAffTbl, buf1) != NULL &&
-		Tcl_FindHashEntry(tablePtr->spanAffTbl, buf2) != NULL) {
-		borders--;
-		*col = -1;
+	    entryPtr1 = Tcl_FindHashEntry(tablePtr->spanAffTbl, buf1);
+	    entryPtr2 = Tcl_FindHashEntry(tablePtr->spanAffTbl, buf2);
+	    if (entryPtr1 != NULL && entryPtr2 != NULL) {
+		if ((val = (char *) Tcl_GetHashValue(entryPtr1)) != NULL) {
+		    strcpy(buf1, val);
+		}
+		if ((val = (char *) Tcl_GetHashValue(entryPtr2)) != NULL) {
+		    strcpy(buf2, val);
+		}
+		if (strcmp(buf1, buf2) == 0) {
+		    borders--;
+		    *col = -1;
+		}
 	    }
 	}
     }
@@ -428,10 +481,15 @@ TableGetCellValue(Table *tablePtr, int r, int c)
     TableMakeArrayIndex(r, c, buf);
 
     if (tablePtr->caching) {
-	/* if we are caching, let's see if we have the value cached */
+	/*
+	 * If we are caching, let's see if we have the value cached
+	 */
 	entryPtr = Tcl_CreateHashEntry(tablePtr->cache, buf, &new);
 	if (!new) {
 	    result = (char *) Tcl_GetHashValue(entryPtr);
+	    if (result == NULL) {
+		result = "";
+	    }
 	    goto VALUE;
 	}
     }
@@ -446,18 +504,22 @@ TableGetCellValue(Table *tablePtr, int r, int c)
 	    Tcl_BackgroundError(interp);
 	    TableInvalidateAll(tablePtr, 0);
 	} else {
-	    result = Tcl_GetResult(interp);
+	    result = (char *) Tcl_GetStringFromObj(Tcl_GetObjResult(interp),NULL);
 	}
 	/* Tcl_FreeResult(interp); not support for perl/tk */
     } else if (tablePtr->arrayVar) {
-	result = LangString(Tcl_GetVar2(interp, tablePtr->arrayVar, buf, TCL_GLOBAL_ONLY));
+	result = Tcl_GetString(Tcl_ObjGetVar2(interp, tablePtr->arrayVar, Tcl_NewStringObj(buf,-1),
+	        TCL_GLOBAL_ONLY));
     }
     if (result == NULL)
 	result = "";
     if (tablePtr->caching && entryPtr != NULL) {
-	/* if we are caching, make sure we cache the returned value */
-	/* entryPtr will have been set from above, but check to make sure
-	 * someone didn't change caching during -command evaluation */
+	/*
+	 * If we are caching, make sure we cache the returned value
+	 *
+	 * entryPtr will have been set from above, but check to make sure
+	 * someone didn't change caching during -command evaluation.
+	 */
 	char *val;
 	val = (char *)ckalloc(strlen(result)+1);
 	strcpy(val, result);
@@ -499,7 +561,6 @@ VALUE:
     return (result?result:"");
 }
 
-extern void	TableAddFlash _ANSI_ARGS_((Table *tablePtr, int row, int col));
 /*
  *----------------------------------------------------------------------
  *
@@ -531,7 +592,7 @@ TableSetCellValue(Table *tablePtr, int r, int c, char *value)
 	return TCL_OK;
     }
     if (tablePtr->command && tablePtr->useCmd) {
-	if (LangDoCallback(interp, tablePtr->command, 1, 4, "%d %d %d %s",1,r,c, value) == TCL_ERROR) {
+	if (LangDoCallback(interp, tablePtr->command, 1, 4, "%d %d %d %_",1,r,c, LangStringArg(value)) == TCL_ERROR) {
 	    /* An error resulted.  Prevent further triggering of the command
 	     * and set up the error message. */
 	    tablePtr->useCmd = 0;
@@ -553,7 +614,7 @@ TableSetCellValue(Table *tablePtr, int r, int c, char *value)
 	    /* Tcl_UnsetVar2(interp, LangString(Tcl_GetVar(interp, tablePtr->arrayVar, TCL_GLOBAL_ONLY)), buf, TCL_GLOBAL_ONLY); */
 	    /* Replaced with This (defined in tkTable.xs) */
 	    tkTableUnsetElement(tablePtr->arrayVar, buf);
-	} else if (Tcl_SetVar2(interp, tablePtr->arrayVar, buf, value,
+	} else if (Tcl_ObjSetVar2(interp, tablePtr->arrayVar, Tcl_NewStringObj(buf,-1), Tcl_NewStringObj(value,-1),
 			       TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) == NULL) {
 	    code = TCL_ERROR;
 	}
@@ -567,9 +628,13 @@ TableSetCellValue(Table *tablePtr, int r, int c, char *value)
 	int new;
 	char *val;
 
+	entryPtr = Tcl_CreateHashEntry(tablePtr->cache, buf, &new);
+	if (!new) {
+	    val = (char *) Tcl_GetHashValue(entryPtr);
+	    if (val) ckfree(val);
+	}
 	val = (char *)ckalloc(strlen(value)+1);
 	strcpy(val, value);
-	entryPtr = Tcl_CreateHashEntry(tablePtr->cache, buf, &new);
 	Tcl_SetHashValue(entryPtr, val);
 	flash = 1;
     }
@@ -584,6 +649,96 @@ TableSetCellValue(Table *tablePtr, int r, int c, char *value)
     return TCL_OK;
 }
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * TableMoveCellValue --
+ *	To move cells faster on delete/insert line or col when cache is on
+ *	and variable, command is off.
+ *	To avoid another call to TableMakeArrayIndex(r, c, buf),
+ *	we optionally provide the buffers.
+ *	outOfBounds means we will just set the cell value to ""
+ *
+ * Results:
+ *	Returns TCL_ERROR or TCL_OK, depending on whether an error
+ *	occured during set (ie: during evaluation of -command).
+ *
+ * Side effects:
+ *	If the value is NULL (empty string), it will be unset from
+ *	an array rather than set to the empty string.
+ *
+ *----------------------------------------------------------------------
+ */
+int
+TableMoveCellValue(Table *tablePtr, int fromr, int fromc, char *frombuf,
+	int tor, int toc, char *tobuf, int outOfBounds)
+{
+    int new;
+    char *result = NULL;
+    Tcl_Interp *interp = tablePtr->interp;
+
+    if (outOfBounds) {
+	return TableSetCellValue(tablePtr, tor, toc, "");
+    }
+
+    if (tablePtr->caching && (!(tablePtr->command && tablePtr->useCmd))) {
+	Tcl_HashEntry *entryPtr;
+	/*
+	 * if we are caching, let's see if we have the value cached
+	 */
+	entryPtr = Tcl_CreateHashEntry(tablePtr->cache, frombuf, &new);
+	if (!new) {
+	    char *val;
+	    result = (char *) Tcl_GetHashValue(entryPtr);
+	    /*
+	     * we set tho old value to NULL
+	     */
+	    Tcl_SetHashValue(entryPtr, NULL);
+
+	    /*
+	     * set the destination to the source pointer without new mallocing!
+	     */
+	    entryPtr = Tcl_CreateHashEntry(tablePtr->cache, tobuf, &new);
+	    /*
+	     * free old value
+	     */
+	    if (!new) {
+		val = (char *) Tcl_GetHashValue(entryPtr);
+		if (val) ckfree(val);
+	    }
+	    Tcl_SetHashValue(entryPtr, result);
+	    if (tablePtr->arrayVar) {
+		/*
+		 * first, delete from var.
+		 */
+		/* perltk not supported  */
+		/*Tcl_UnsetVar2(interp, tablePtr->arrayVar, frombuf,
+			TCL_GLOBAL_ONLY);
+		*/
+		/* Replaced with This (defined in tkTable.xs) */
+		tkTableUnsetElement(tablePtr->arrayVar, frombuf);
+		/*
+		 * Warning: checking for \0 as the first char could invalidate
+		 * allowing it as a valid first char
+		 */
+		if (Tcl_ObjSetVar2(interp, tablePtr->arrayVar, Tcl_NewStringObj(tobuf,-1), Tcl_NewStringObj(result, -1),
+			TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) == NULL) {
+		    return TCL_ERROR;
+		}
+	    }     
+      
+
+	    return TCL_OK;
+	}
+    }
+    /*
+     * We have to do it the old way
+     */
+    return TableSetCellValue(tablePtr, tor, toc,
+	    TableGetCellValue(tablePtr, fromr, fromc));
+
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -618,7 +773,7 @@ TableGetIcursor(Table *tablePtr, char *arg, int *posn)
     } else if (strcmp(arg, "insert") == 0) {
 	tmp = tablePtr->icursor;
     } else {
-	if (Tcl_GetInt(tablePtr->interp, LangStringArg(arg), &tmp) != TCL_OK) {
+	if (Tcl_GetIntFromObj(tablePtr->interp, LangStringArg(arg), &tmp) != TCL_OK) {
 	    return TCL_ERROR;
 	}
 	CONSTRAIN(tmp, 0, len);
@@ -651,11 +806,11 @@ TableGetIcursor(Table *tablePtr, char *arg, int *posn)
  */
 int
 TableGetIndex(tablePtr, str, row_p, col_p)
-     register Table *tablePtr;	/* Table for which the index is being
+    register Table *tablePtr;	/* Table for which the index is being
 				 * specified. */
-     char *str;			/* Symbolic specification of cell in table. */
-     int *row_p;		/* Where to store converted row. */
-     int *col_p;		/* Where to store converted col. */
+    char *str;			/* Symbolic specification of cell in table. */
+    int *row_p;		/* Where to store converted row. */
+    int *col_p;		/* Where to store converted col. */
 {
     int r, c, len = strlen(str);
     char dummy;
@@ -711,13 +866,20 @@ TableGetIndex(tablePtr, str, row_p, col_p)
 	r = tablePtr->topRow+tablePtr->rowOffset;
 	c = tablePtr->leftCol+tablePtr->colOffset;
     } else if (strncmp(str, "bottomright", len) == 0) {	/* bottomright */
+	/*
+	 * FIX: Should this avoid spans, or consider them in the bottomright?
+	 tablePtr->flags |= AVOID_SPANS;
+	 tablePtr->flags &= ~AVOID_SPANS;
+	 */
 	TableGetLastCell(tablePtr, &r, &c);
 	r += tablePtr->rowOffset;
 	c += tablePtr->colOffset;
     } else {
     IndexError:
 	Tcl_AppendStringsToObj(Tcl_GetObjResult(tablePtr->interp),
-			       "bad table index \"", str, "\"", (char *)NULL);
+		"bad table index \"", str, "\": must be active, anchor, end, ",
+		"origin, topleft, bottomright, @x,y, or <row>,<col>",
+		(char *)NULL);
 	return TCL_ERROR;
     }
 
@@ -751,7 +913,6 @@ Table_SetCmd(ClientData clientData, register Tcl_Interp *interp,
     register Table *tablePtr = (Table *)clientData;
     int row, col, len, i, j, max;
     char *str;
-    Tcl_Obj *resultPtr;
 
     /* sets any number of tags/indices to a given value */
     if (objc < 3) {
@@ -766,9 +927,9 @@ Table_SetCmd(ClientData clientData, register Tcl_Interp *interp,
 	return TCL_OK;
     }
 
-    resultPtr = Tcl_GetObjResult(interp);
     str = Tcl_GetStringFromObj(objv[2], &len);
     if (strncmp(str, "row", len) == 0 || strncmp(str, "col", len) == 0) {
+	Tcl_Obj *resultPtr = Tcl_GetObjResult(interp);
 	/* set row index list ?index list ...? */
 	if (objc < 4) {
 	    goto CMD_SET_USAGE;
@@ -846,8 +1007,12 @@ Table_SetCmd(ClientData clientData, register Tcl_Interp *interp,
 	if (TableGetIndexObj(tablePtr, objv[2], &row, &col) != TCL_OK) {
 	    return TCL_ERROR;
 	} else {
-	    Tcl_SetStringObj(resultPtr,
-			     TableGetCellValue(tablePtr, row, col), -1);
+	    /*
+	     * Cannot use Tcl_GetObjResult here because TableGetCellValue
+	     * can corrupt the resultPtr.
+	     */
+	    Tcl_SetObjResult(interp, Tcl_NewStringObj(
+		TableGetCellValue(tablePtr, row, col),-1));
 	}
     } else {
 	/* set index val ?index val ...? */
@@ -888,20 +1053,24 @@ Table_SetCmd(ClientData clientData, register Tcl_Interp *interp,
  *--------------------------------------------------------------
  */
 static int
-Table_SpanSet(register Table *tablePtr, int row, int col, int rs, int cs)
+Table_SpanSet(register Table *tablePtr, int urow, int ucol, int rs, int cs)
 {
     Tcl_Interp *interp = tablePtr->interp;
     int i, j, new, ors, ocs, result = TCL_OK;
+    int row, col;
     Tcl_HashEntry *entryPtr;
     Tcl_HashSearch search;
     char *dbuf, buf[INDEX_BUFSIZE], cell[INDEX_BUFSIZE], span[INDEX_BUFSIZE];
 
-    TableMakeArrayIndex(row, col, cell);
+    row = urow - tablePtr->rowOffset;
+    col = ucol - tablePtr->colOffset;
+
+    TableMakeArrayIndex(urow, ucol, cell);
 
     if (tablePtr->spanTbl == NULL) {
-	tablePtr->spanTbl = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
+	tablePtr->spanTbl = (Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
 	Tcl_InitHashTable(tablePtr->spanTbl, TCL_STRING_KEYS);
-	tablePtr->spanAffTbl = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
+	tablePtr->spanAffTbl = (Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
 	Tcl_InitHashTable(tablePtr->spanAffTbl, TCL_STRING_KEYS);
     }
 
@@ -920,13 +1089,13 @@ Table_SpanSet(register Table *tablePtr, int row, int col, int rs, int cs)
      * title cells must not expand beyond the titles
      * other cells can't expand negatively into title area
      */
-    if ((row-tablePtr->rowOffset < tablePtr->titleRows) &&
-	(row-tablePtr->rowOffset+rs >= tablePtr->titleRows)) {
-	rs = tablePtr->titleRows-(row-tablePtr->rowOffset)-1;
+    if ((row < tablePtr->titleRows) &&
+	(row + rs >= tablePtr->titleRows)) {
+	rs = tablePtr->titleRows - row - 1;
     }
-    if ((col-tablePtr->colOffset < tablePtr->titleCols) &&
-	(col-tablePtr->colOffset+cs >= tablePtr->titleCols)) {
-	cs = tablePtr->titleCols-(col-tablePtr->colOffset)-1;
+    if ((col < tablePtr->titleCols) &&
+	(col + cs >= tablePtr->titleCols)) {
+	cs = tablePtr->titleCols - col - 1;
     }
     rs = MAX(0, rs);
     cs = MAX(0, cs);
@@ -937,8 +1106,8 @@ Table_SpanSet(register Table *tablePtr, int row, int col, int rs, int cs)
 	TableParseArrayIndex(&ors, &ocs, (char *)Tcl_GetHashValue(entryPtr));
 	ckfree((char *) Tcl_GetHashValue(entryPtr));
 	Tcl_DeleteHashEntry(entryPtr);
-	for (i = row; i <= row+ors; i++) {
-	    for (j = col; j <= col+ocs; j++) {
+	for (i = urow; i <= urow+ors; i++) {
+	    for (j = ucol; j <= ucol+ocs; j++) {
 		TableMakeArrayIndex(i, j, buf);
 		entryPtr = Tcl_FindHashEntry(tablePtr->spanAffTbl, buf);
 		if (entryPtr != NULL) {
@@ -953,14 +1122,14 @@ Table_SpanSet(register Table *tablePtr, int row, int col, int rs, int cs)
     }
 
     /* calc to make sure that span is OK */
-    for (i = row; i <= row+rs; i++) {
-	for (j = col; j <= col+cs; j++) {
+    for (i = urow; i <= urow+rs; i++) {
+	for (j = ucol; j <= ucol+cs; j++) {
 	    TableMakeArrayIndex(i, j, buf);
 	    entryPtr = Tcl_FindHashEntry(tablePtr->spanAffTbl, buf);
 	    if (entryPtr != NULL) {
 		/* Something already spans here */
 		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
-				       "cannot overlap already spanning cell ",
+				       "cannot overlap already spanned cell ",
 				       buf, (char *) NULL);
 		result = TCL_ERROR;
 		rs = ors;
@@ -1009,19 +1178,17 @@ Table_SpanSet(register Table *tablePtr, int row, int col, int rs, int cs)
     Tcl_SetHashValue(entryPtr, dbuf);
     dbuf = Tcl_GetHashKey(tablePtr->spanTbl, entryPtr);
     /* Set other affected cells */
-    EmbWinUnmap(tablePtr, row-tablePtr->rowOffset, row+rs-tablePtr->rowOffset,
-		col-tablePtr->colOffset, col+cs-tablePtr->colOffset);
-    for (i = row; i <= row+rs; i++) {
-	for (j = col; j <= col+cs; j++) {
+    EmbWinUnmap(tablePtr, row, row + rs, col, col + cs);
+    for (i = urow; i <= urow+rs; i++) {
+	for (j = ucol; j <= ucol+cs; j++) {
 	    TableMakeArrayIndex(i, j, buf);
 	    entryPtr = Tcl_CreateHashEntry(tablePtr->spanAffTbl, buf, &new);
-	    if (!(i == row && j == col)) {
+	    if (!(i == urow && j == ucol)) {
 		Tcl_SetHashValue(entryPtr, (char *) dbuf);
 	    }
 	}
     }
-    TableRefresh(tablePtr, row-tablePtr->rowOffset,
-		 col-tablePtr->colOffset, CELL);
+    TableRefresh(tablePtr, row, col, CELL);
     return result;
 }
 
@@ -1151,7 +1318,7 @@ Table_HiddenCmd(ClientData clientData, register Tcl_Interp *interp,
 				     Tcl_GetHashKey(tablePtr->spanAffTbl,
 						    entryPtr));
 	}
-	span = LangString(TableCellSort(tablePtr, Tcl_DStringValue(&cells)));
+	span = Tcl_GetString(TableCellSort(tablePtr, Tcl_DStringValue(&cells)));
 	if (span != NULL) {
 	    Tcl_SetResult(interp, span, TCL_DYNAMIC);
 	}
